#cython: language_level=3, boundscheck=False, overflowcheck=False
#    profile=True, linetrace=True
from libc.stdlib cimport malloc, free, realloc

cdef int char_num(byte_t *src):
    cdef int ret = 0
    cdef int char_ = 0
    cdef int char_byte_num = 0
    while True:
        src = iter_unicode(src, &char_, &char_byte_num)
        if src == NULL:
            return -1
        if char_ == 0:
            break
        ret += 1
    return ret


cdef int fill_char_info(byte_t *src, int *char_idx_of_byte, int *chars, int *char_offsets):
    cdef int ret = 0
    cdef int char_ = 0
    cdef int char_byte_num = 0
    cdef int offset = 0
    cdef int i
    while True:
        src = iter_unicode(src, &char_, &char_byte_num)
        if src == NULL:
            return -1
        if char_ == 0:
            break
        char_offsets[ret] = offset
        chars[ret] = char_
        for i in range(char_byte_num):
            char_idx_of_byte[offset + i] = ret
        offset += char_byte_num
        ret += 1
    char_idx_of_byte[offset] = ret
    char_offsets[ret] = offset
    return ret



#cython: language_level=3, boundscheck=False, overflowcheck=False
#    , profile=True, linetrace=True

from libc.stdlib cimport malloc, free, realloc
from libc.string cimport memcpy 
from libc.stdio cimport FILE, fopen, fwrite, fclose
from libcpp.string cimport string
from libc.stdint cimport uint32_t
from cython cimport typeof
from .util cimport check_buffer
from cpython.buffer cimport PyObject_GetBuffer, PyObject_CheckBuffer, PyBuffer_Release, PyBuffer_GetPointer, Py_buffer, PyBUF_WRITABLE, PyBUF_SIMPLE
from .version import magic_number, legacy_magic_number

def new_object(obj):
    return obj.__new__(obj)

cdef extern from "<algorithm>" namespace "std" nogil:
    void reverse[T](T a, T b);

cdef int CHILD_NUM_MASK = (1 << 9) -1
cdef int END_MASK = 1 << 9
cdef int value_limit = (1 << 31) - 1

cdef inline int ignore_case_byte_index_mapping(ignore_case_alignment align, int byte_idx):
    if align is None:
        return byte_idx
    cdef int char_offset = align.lowercase.char_idx_of_byte[byte_idx]
    char_offset = align.lowercase_char_index_mapping[char_offset]
    cdef int ret = align.original.char_offsets[char_offset]
    return ret


cdef class Trie(object):

    """Trie
    Attributes:
        ignore_case (bool): if the trie should be case sensitive or not. 
            if it's true, all inserted keys will be converted to lowercase.
            Defaults to False
        ordered     (bool): the child list of each node should be ordered.
            Defaults to False
    Examples:
        >>> # import module
        >>> from cyac import Trie
        >>> # Create an object of Trie
        >>> trie = Trie()
        >>> # add keywords
        >>> keys = [u'abc', u'cde', u'efg']
        >>> for key in keys:
        >>>     trie.insert(key)
    """

    cdef inline void _node_init(self, Node* n, int value, int check):
        n.value = value
        n.check = check
        n.flags = 0
        n.child = 0
        n.sibling = 0

    cdef inline int _node_child_num(self, Node *n):
        return n.flags
    
    cdef inline void _node_set_child_num(self, Node *n, int val):
        n.flags = val

    cdef inline int _node_base(self, Node *n):
        return -(n.value + 1)

    cdef inline bool _node_is_child(self, Node *n, int par):
        return n.check == par

    cdef inline byte_t* _node_child_ptr(self, Node *n):
        return &n.child

    cdef inline byte_t* _node_sibling_ptr(self, Node *n):
        return &n.sibling

    
    cdef inline void _block_init(self, Block* block, int prev, int next_, int trial, int ehead, int num, int reject):
        block.prev = prev
        block.next_ = next_
        block.trial = trial
        block.ehead = ehead
        block.num = num
        block.reject = reject

    def arr_size(self):
        return self.array_size

    property key_num:
        def __get__(self):
            return self.key_num
    
    property size:
        def __get__(self):
            return self.key_num

    def __len__(self):
        return self.key_num
    
    cdef inline Node* _node(self, int nid):
        return self.array + nid

    def __cinit__(self, bool ignore_case = False, bool ordered = False):
        cdef int i
        self.last_remove_leaf = value_limit
        self.ignore_case = ignore_case
        self.ordered = ordered
        self.capacity = 256
        self.key_num = 0
        self.leaf_size = 0
        self.key_capacity = self.capacity
        self.leafs = <int*> malloc(self.key_capacity * sizeof(int))
        self.array = <Node*> malloc(self.capacity * sizeof(Node))
        self.array_size = self.capacity
        self.blocks = <Block*> malloc((self.capacity >> 8) * sizeof(Block))
        self.max_trial = 1
        self._node_init(self.array, -1, -1)
        for i in range(1, 256):
            self._node_init(&self.array[i], -(i - 1), -(i + 1))
        self.array[1].value = -255
        self.array[255].check = -1
        self._block_init(self.blocks, 0, 0, 0, 1, 256, 257)
        for i in range(0, 257):
            self.reject[i] = i + 1
        
        self.bheadF = 0
        self.bheadC = 0
        self.bheadO = 0
        self.buff = NULL
    
    cdef inline int _get(self, byte_t *key, int key_size, int from_, int start):
        cdef int pos, value, to
        for pos in range(start, key_size):
            value = self.array[from_].value
            if value >= 0 and value != value_limit:
                to = self.follow(from_, 0)
                self.array[to].value = value
                self.leafs[value] = to
            from_ = self.follow(from_, key[pos])
        if self.array[from_].value < 0:
            return self.follow(from_, 0)
        else:
            return from_

    cdef inline int follow(self, int from_, byte_t label):
        cdef int base = self._node_base(&self.array[from_])
        cdef int to = base ^ label
        cdef bool has_child
        if base < 0  or self.array[to].check < 0:
            has_child = base >= 0 and self.array[base ^ self.array[from_].child].check == from_
            to = self.pop_enode(base, label, from_)
            self.push_sibling(from_, to^label, label, has_child)
        elif self.array[to].check != from_:
            to = self.resolve(from_, base, label)
        return to
    
    cdef inline void pop_block(self, int bi, int* head_in, bool last):
        if last:
            head_in[0] = 0
        else:
            self.blocks[self.blocks[bi].prev].next_ = self.blocks[bi].next_
            self.blocks[self.blocks[bi].next_].prev = self.blocks[bi].prev
            if bi == head_in[0]:
                head_in[0] = self.blocks[bi].next_
    
    cdef inline void push_block(self, int bi, int* head_out, bool empty):
        cdef Block *tail_out
        cdef Block *b = self.blocks + bi
        if empty:
            head_out[0] = bi
            b.prev = bi
            b.next_ = bi
        else:
            tail_out = &self.blocks[head_out[0]]
            b.prev = tail_out[0].prev
            b.next_ = head_out[0]
            self.blocks[tail_out.prev].next_ = bi
            head_out[0] = bi
            tail_out[0].prev = bi

    cdef int add_block(self):
        cdef int i
        if self.array_size == self.capacity:
            self.capacity *= 2
            self.blocks = <Block*> realloc(self.blocks, sizeof(Block) * (self.capacity >> 8))
            self.array = <Node*> realloc(self.array, sizeof(Node) * self.capacity)
        self._block_init(&self.blocks[self.array_size >> 8], 0,0,0, self.array_size, 256, 257)
        for i in range(0, 256):
            self._node_init(
                &self.array[self.array_size + i],
                 -(((i + 255) & 255) + self.array_size), 
                 -(((i + 1) & 255) + self.array_size))
        self.push_block(self.array_size >> 8, &self.bheadO, self.bheadO == 0)
        self.array_size += 256
        return (self.array_size >> 8) - 1

    cdef inline void transfer_block(self, int bi, int* head_in, int* head_out):
        self.pop_block(bi, head_in, bi == self.blocks[bi].next_)
        self.push_block(bi, head_out, head_out[0] == 0 and self.blocks[bi].num != 0)

    cdef inline int pop_enode(self, int base, byte_t label, int from_):
        cdef int e = self.find_place() if base < 0 else (base ^ label)
        cdef int bi = e >> 8
        cdef Node *n = &self.array[e]
        cdef Block *b = &self.blocks[bi]
        b.num -= 1
        if b.num == 0:
            if bi != 0:
                self.transfer_block(bi, &self.bheadC, &self.bheadF)
        else:
            self.array[-n.value].check = n.check
            self.array[-n.check].value = n.value
            if e == b.ehead:
                b.ehead = -n.check
            if bi != 0 and b.num == 1 and b.trial != self.max_trial:
                self.transfer_block(bi, &self.bheadO, &self.bheadC)
        n.value = value_limit
        n.check = from_
        if base < 0:
            self.array[from_].value = -(e^label) - 1
        return e

    cdef void push_enode(self, int e):
        cdef Node *e_ptr = &self.array[e]
        cdef int bi = e >> 8
        cdef Block *b = &self.blocks[bi]
        cdef int prev, next_
        cdef Node *prev_ptr
        b.num += 1
        if b.num == 1:
            b.ehead = e
            e_ptr[0].value = -e
            e_ptr[0].check = -e
            if bi != 0:
                self.transfer_block(bi, &self.bheadF, &self.bheadC)
        else:
            prev = b.ehead
            prev_ptr = &self.array[prev]
            next_ = -prev_ptr[0].check
            e_ptr[0].value = -prev
            e_ptr[0].check = -next_
            prev_ptr[0].check = -e
            self.array[next_].value = -e
            if b.num == 2 or b.trial == self.max_trial:
                if bi != 0:
                    self.transfer_block(bi, &self.bheadC, &self.bheadO)
            b.trial = 0
        if b.reject < self.reject[b.num]:
            b.reject = self.reject[b.num]
        e_ptr[0].child = 0
        e_ptr[0].sibling = 0
        e_ptr[0].flags = 0
    
    cdef inline void push_sibling(self, int from_, int base, byte_t label, bool has_child):
        cdef Node *from_ptr = &self.array[from_]
        cdef byte_t *child_ptr = self._node_child_ptr(from_ptr)
        cdef bool keep_order = (label > child_ptr[0]) if self.ordered else (child_ptr[0] == 0)
        if has_child and keep_order:
            child_ptr = self._node_sibling_ptr(&self.array[base ^ child_ptr[0]])
            while self.ordered and child_ptr[0] != 0 and child_ptr[0] < label:
                c = self._node_sibling_ptr(&self.array[base ^ child_ptr[0]])
        
        self.array[base^label].sibling = child_ptr[0]
        child_ptr[0] = label
        self._node_set_child_num(from_ptr, self._node_child_num(from_ptr) + 1)

    cdef inline void pop_sibling(self, int from_, byte_t label):
        cdef Node *from_ptr = &self.array[from_]
        cdef int base = self._node_base(&from_ptr[0])
        cdef byte_t *child_ptr = self._node_child_ptr(from_ptr)
        while child_ptr[0] != label:
            child_ptr = self._node_sibling_ptr(&self.array[base ^ child_ptr[0]])
        child_ptr[0] = self.array[base ^ child_ptr[0]].sibling
        self._node_set_child_num(from_ptr, self._node_child_num(from_ptr) - 1)


    cdef inline bool consult(self, Node *nref, Node *pref):
        return self._node_child_num(nref) < self._node_child_num(pref)

    cdef bool has_label(self, int id_, byte_t label):
        return self.child(id_, label) >= 0

    cdef inline int child(self, int id_, byte_t label):
        cdef Node *parent_ptr = &self.array[id_]
        cdef int base = self._node_base(&parent_ptr[0])
        cdef int cid = base ^ label
        # print("cid[%s][%s] = %s, check[%s] = %s" % (id_, label, cid, cid, self.array[cid].check))
        if cid < 0  or  cid >= self.array_size  or not self._node_is_child(&self.array[cid], id_):
            return -1 
        return cid

    cdef inline int children(self, int id_, byte_t *labels, int *children_arr, int first_n):
        cdef Node *parent_ptr = &self.array[id_]
        cdef int base = self._node_base(&parent_ptr[0])
        cdef byte_t s = parent_ptr[0].child
        if s == 0 and base > 0:
            s = self.array[base].sibling
        cdef int num = 0
        while s != 0:
            to = base ^ s
            if to < 0:
                break
            labels[num] = s
            children_arr[num] = to
            num += 1
            if num >= first_n:
                break
            s = self.array[to].sibling
        return num

    
    cdef inline int sibling(self, int to):
        if to < 0:
            return -1
        cdef int base = self._node_base(&self.array[self.array[to].check])
        cdef byte_t s = self.array[to].sibling
        if s == 0:
            return -1
        to = base ^ s
        return to

    # 返回children个数
    cdef int set_child(self, int base, byte_t c, byte_t label, bool append_label, byte_t *children):
        cdef int idx = 0
        if c == 0:
            children[idx] = c
            idx += 1
            c = self.array[base ^ c].sibling
        
        if self.ordered:
            while c != 0 and c <= label:
                children[idx] = c
                idx += 1
                c = self.array[base ^c].sibling
        
        if append_label:
            children[idx] = label
            idx += 1
        
        while c != 0:
            children[idx] = c
            idx += 1
            c = self.array[base ^ c].sibling
        return idx

    cdef inline int find_place(self):
        if self.bheadC != 0:
            return self.blocks[self.bheadC].ehead 
        if self.bheadO != 0:
            return self.blocks[self.bheadO].ehead 
        return self.add_block() << 8

    cdef inline int find_places(self, byte_t *child, int child_num):
        cdef int bi = self.bheadO
        cdef int bz, e, i
        cdef Block *b
        if bi != 0:
            bz = self.blocks[self.bheadO].prev
            nc = child_num
            while True:
                b = &self.blocks[bi]
                if b.num >= nc  and  nc < b.reject:
                    e = b.ehead
                    while True:
                        base = e ^ child[0]
                        i = 0
                        for i in range(0, child_num):
                            c = child[i]
                            if not self.array[base ^ c].check < 0:
                                break
                            if i == child_num - 1:
                                b.ehead = e
                                return e
                        e = - self.array[e].check
                        if e == b.ehead:
                            break
                b.reject = nc
                if b.reject < self.reject[b.num]:
                    self.reject[b.num] = b.reject
                bin_ = b.next_
                b.trial += 1
                if b.trial == self.max_trial:
                    self.transfer_block(bi, &self.bheadO, &self.bheadC)
                if bi == bz:
                    break
                bi = bin_
        return self.add_block() << 8

    cdef inline int resolve(self, int from_n, int base_n, byte_t label_n):
        cdef int to_pn = base_n ^ label_n
        cdef int from_p = self.array[to_pn].check
        cdef Node *from_p_ptr = &self.array[from_p]
        cdef Node *from_n_ptr = &self.array[from_n]
        cdef int base_p = self._node_base(&self.array[from_p])
        # cdef Node *to_ptr 
        cdef Node *to_ptr_
        cdef Node *ptr
        cdef Node *n_
        cdef Node *from_ptr
        cdef bool flag = self.consult(from_n_ptr, from_p_ptr)
        cdef byte_t children[256]
        cdef int from_, i, child_num, base
        if flag:
            child_num = self.set_child(base_n, from_n_ptr[0].child, label_n, True, children)
        else:
            child_num = self.set_child(base_p, from_p_ptr[0].child, 255, False, children)
        base = self.find_place() if child_num == 1 else self.find_places(children, child_num)
        base ^= children[0]
        if flag:
            from_ = from_n
            from_ptr = &self.array[from_n]
            base_ = base_n
        else:
            from_ = from_p
            from_ptr = &self.array[from_p]
            base_ = base_p
        if flag  and  children[0] == label_n:
            from_ptr[0].child = label_n
        from_ptr[0].value = -base - 1
        for i in range(0, child_num):
            chl = children[i]
            to = self.pop_enode(base, chl, from_)
            to_ = base_ ^ chl
            n = &self.array[to]
            if i == child_num - 1:
                n.sibling = 0
            else:
                n.sibling = children[i + 1]
            if flag  and  to_ == to_pn:
                continue
            n_ = &self.array[to_]
            n.value = n_.value
            if n_.value >= 0  and  n_.value != value_limit:
                self.leafs[n_.value] = to
            n.flags = n_.flags
            if n.value < 0  and  chl != 0:
                c = self.array[to_].child
                self.array[to].child = c
                ptr = &self.array[self._node_base(n) ^ c]
                ptr[0].check = to
                c = ptr[0].sibling
                while c != 0:
                    ptr = &self.array[self._node_base(n)^ c]
                    ptr[0].check = to
                    c = ptr[0].sibling
            
            if not flag and to_ == from_n:
                from_n = to
            if not flag and to_ == to_pn:
                self.push_sibling(from_n, to_pn ^ label_n, label_n, True)
                to_ptr_ = &self.array[to_]
                to_ptr_[0].child = 0
                n_.value = value_limit
                n_.check = from_n
            else:
                self.push_enode(to_)
        if flag:
            return base ^ label_n
        else:
            return to_pn
    
    cdef inline int jump(self, byte_t byte, int from_):
        from_ptr = &self.array[from_]
        if from_ptr[0].value >= 0:
            return -1
        to = self._node_base(&from_ptr[0]) ^ byte
        if self.array[to].check != from_:
            return -1
        return to

    cdef inline int jump_bytes(self, byte_t *bytes_, int byte_num, int from_):
        cdef int i
        cdef byte_t byte
        for i in range(byte_num):
            byte = bytes_[i]
            from_ = self.jump(byte, from_)
            if from_ < 0:
                return -1
        return from_

    cdef inline int jump_uchar(self, xstring text, int char_idx, int from_):
        cdef byte_t *start = text.bytes_ + text.char_offsets[char_idx]
        cdef int num = text.char_offsets[char_idx+1] - text.char_offsets[char_idx]
        return self.jump_bytes(start, num, from_)


    cdef bytes substring(self, int id_, int start_id): # end_id is included, start_id is excluded
        cdef stringbuf sb
        cdef int from_
        cdef byte_t chr_
        while id_ > 0 and id_ != start_id:
            from_ = self.array[id_].check
            if from_ < 0:
                return None
            chr_ = self._node_base(&self.array[from_]) ^ id_
            if chr_ >= 256:
                return None
            if chr_ != 0:
                sb.put(chr_)
            id_ = from_
        if id_ != start_id:
            return None
        cdef string ret = sb.to_string()
        reverse(ret.begin(), ret.end())
        return ret

    cdef bytes key(self, int id_):
        return self.substring(id_, 0)

    cdef inline int value(self, int id_):
        cdef Node *ptr = &self.array[id_]
        cdef int val = ptr[0].value
        if val >= 0:
            return val
        cdef int to = self._node_base(ptr)
        cdef Node *to_ptr = &self.array[to]
        if to_ptr[0].check == id_  and  to_ptr[0].value >= 0  and  to_ptr[0].value != value_limit:
            return to_ptr[0].value
        return -1

    cdef inline bool has_value(self, int id_):
        cdef Node *ptr = &self.array[id_]
        cdef int val = ptr[0].value
        if val >= 0:
            return True
        cdef int to = self._node_base(ptr)
        cdef Node *to_ptr = &self.array[to]
        if to_ptr[0].check == id_  and  to_ptr[0].value >= 0:
            return True
        return False

    def __contains__(self, unicode t):
        return self.get(t) >= 0

    def __iter__(self):
        cdef int id_
        cdef int lnode
        for id_ in range(0,self.leaf_size):
            lnode = self.leafs[id_]
            if lnode >= 0:
                yield id_

    def __getitem__(self, sid):
        """
        get key by id, or get id by key, raise exception if cannot find.
        Args:
            sid (unicode | int): key or id

        Examples:
            >>> trie["python"]
        Raises:
            AttributeError: If `sid` is not unicode or int
        """
        cdef int lnode
        if isinstance(sid, int):
            if sid >= self.leaf_size:
                raise AttributeError("index out of range: %d >= %d" % (sid, self.leaf_size))
            lnode = self.leafs[sid]
            if lnode < 0:
                raise AttributeError("Cannot find with id: %d" % sid)
            bs = self.key(lnode)
            bs = bs.decode("utf8")
            return bs
        elif isinstance(sid, unicode):
            ret = self.get(sid)
            if ret < 0:
                raise Exception("Cannot find: %s" % repr(sid))
            return ret
        raise AttributeError("Index type is not supported")

    cpdef int insert(self, unicode key):
        """insert key into the trie, return the id of this key.
        Args:
            key : string, Cannot be empty.
        Returns:
            id  : int
                The id is continuously increasing. you can use this id to index other values.
                If some key is removed, and a new key is inserted, the old key's id will be assigned to the new key.
        Examples:
            >>> trie.insert("python")
        """
        if self.ignore_case:
            key = key.lower()
        cdef bytes bkey = key.encode("utf8")
        cdef int bkey_len = len(bkey)
        if bkey_len == 0:
            return -1
        cdef byte_t *ckey = bkey
        val = self.get_bytes(ckey, bkey_len)
        if val >= 0:
            return val
        cdef int p = self._get(ckey, bkey_len, 0, 0)
        cdef int id_
        if self.last_remove_leaf != value_limit:
            id_ = self.last_remove_leaf
            self.last_remove_leaf = -(self.leafs[id_] + 1)
        else:
            id_ = self.leaf_size
            if self.leaf_size == self.key_capacity:
                self.key_capacity *= 2
                self.leafs = <int*> realloc(self.leafs, self.key_capacity*sizeof(int))
            self.leaf_size += 1
        cdef Node *p_ptr = &self.array[p]
        p_ptr[0].value = id_
        self.leafs[id_] = p
        self.key_num += 1
        return id_

    cpdef int remove(self, unicode key):
        """
        remove the given key from the trie
        Args:
            key : string
                keyword that you want to remove if it's present.
        Returns:
            id : bool
                the id of key.
                if this key doesn't exist in trie, then return -1.
        Examples:
            >>> trie.remove("python")
        """
        if self.ignore_case:
            key = key.lower()
        cdef Node *from_ptr
        cdef bytes bkey = key.encode("utf8")
        cdef byte_t* cbkey = bkey
        cdef int to = self.jump_bytes(cbkey, len(bkey), 0)
        cdef int base, from_
        cdef byte_t label
        if to < 0:
            return -1
        cdef int vk = self.value(to)
        if vk < 0:
            return -1
        cdef Node *to_ptr = &self.array[to]
        if to_ptr[0].value < 0:
            base = self._node_base(to_ptr)
            if self.array[base].check == to:
                to = base
        while to > 0:
            to_ptr = &self.array[to]
            from_ = to_ptr[0].check
            from_ptr = &self.array[from_]
            base = self._node_base(from_ptr)
            label = (to ^ base)
            if to_ptr[0].sibling != 0  or  from_ptr[0].child != label:
                self.pop_sibling(from_, label)
                self.push_enode(to)
                break
            self.push_enode(to)
            to = from_
        self.key_num -= 1
        self.leafs[vk] = -self.last_remove_leaf - 1
        self.last_remove_leaf = vk
        return vk

    cpdef int get(self, unicode key):
        """
        get id of given key, if it doesn't exist, return -1.
        Args:
            key : string
                keyword that you want to get if it's present.
        Returns:
            id : bool
                the id of key.
                if this key doesn't exist in trie, then return -1.
        Examples:
            >>> trie.get("python")
        """
        if self.ignore_case:
            key = key.lower()
        cdef bytes bkey = key.encode("utf8")
        cdef byte_t* ckey = bkey
        return self.get_bytes(ckey, len(bkey))

    cdef inline int get_bytes(self, byte_t* key, int size):
        cdef int to = self.jump_bytes(key, size, 0)
        if to < 0:
            return -1
        cdef int vk = self.value(to)
        if vk < 0:
            return -1
        return vk


    def prefix(self, unicode s not None):
        """
        return the prefix of given string which is in the trie.
        Args:
            key : string
                keyword that you want to searh
        Iterates:
            prefixes : tuple(id, end_offset)
                s[:end_offset] matches id
        Examples:
            >>> for id_, offset in trie.prefix("python"):
            >>>     print(id_, offset)
        """
        cdef ignore_case_alignment align = None
        cdef xstring xs = xstring(s)
        if self.ignore_case:
            align = ignore_case_alignment(xs)
            xs = align.lowercase
        cdef byte_t b
        cdef int node = 0
        cdef int bi
        for bi in range(xs.byte_num):
            b = xs.bytes_[bi]
            node = self.jump(b, node)
            if node >= 0:
                vk = self.value(node)
                if vk >= 0:
                    yield vk, ignore_case_offset(align, xs, bi) + 1
            else:
                break

    def predict(self, unicode s not None):
        """
        return the string in the trie which starts with given string
        Args:
            key : string
                keyword that you want to searh
        Iterates:
            predicts : id
        Examples:
            >>> for id_ in trie.predict("python"):
            >>>     print(id_)
        """
        cdef ignore_case_alignment align = None
        cdef xstring xs = xstring(s)
        if self.ignore_case:
            align = ignore_case_alignment(xs)
            xs = align.lowercase
        cdef int node = self.jump_bytes(xs.bytes_, xs.byte_num, 0)
        if node < 0:
            return
        cdef int cid, vk, idx
        cdef deque[int] q = deque[int]()
        cdef int child_num = 0
        cdef byte_t children[256]
        cdef int children_nodes[256]
        q.push_back(node)
        while q.size() > 0:
            node = q.front()
            q.pop_front()
            vk = self.value(node)
            if vk >= 0:
                yield vk
            num = self.children(node, children, children_nodes, 256)
            for idx in range(num):
                q.push_back(children_nodes[idx])

    def items(self):
        """
        return all key and id ordered by id
        Iterates:
            items : (unicode, int)
        Examples:
            >>> for key, id_ in trie.items():
            >>>     print(id_, key)
        """
        cdef int id_
        cdef int lnode
        for id_ in range(0,self.leaf_size):
            lnode = self.leafs[id_]
            if lnode >= 0:
                key_ = self.key(lnode)
                yield key_.decode("utf8"), id_


    def __dealloc__(self):
        if self.buff == NULL:
            free(self.leafs)
            free(self.array)
            free(self.blocks)
        else:
            PyBuffer_Release(self.buff)
            free(self.buff)
    
    def match_longest(self, unicode s not None, sep = None):
        """
        extract trie's keys from given string. only return the longest.
        Args:
            s : unicode
            sep : set(int) | None
                If you specify seperators. e.g. set([ord(' ')]), 
                it only matches strings tween seperators.
        Iterates:
            matched: tuple(id, start_offset, end_offset)
        Examples:
            >>> for id_, start_offset, end_offset in trie.match_longest("python", set([ord(" ")])):
            >>>     print(id_, start_offset, end_offset)
        """
        cdef ignore_case_alignment align = None
        cdef xstring xs = xstring(s)
        if self.ignore_case:
            align = ignore_case_alignment(xs)
            xs = align.lowercase
        cdef int offset = 0 # byte offset
        cdef int node = 0
        cdef int bi
        cdef int last_vk
        cdef int last_b
        while offset < xs.byte_num:
            last_vk = -1
            last_b = 0
            if sep is not None:
                while offset < xs.byte_num and xs.chars[xs.char_idx_of_byte[offset]] in sep:
                    offset += xs.char_byte_num(xs.char_idx_of_byte[offset])
                if offset >= xs.byte_num:
                    break
            node = 0
            for bi in range(offset, xs.byte_num):
                b = xs.bytes_[bi]
                node = self.jump(b, node)
                if node >= 0:
                    vk = self.value(node)
                    if vk >= 0:
                        if sep is not None:
                            if bi + 1 < xs.byte_num and xs.chars[xs.char_idx_of_byte[bi+1]] not in sep:
                                continue
                        last_vk = vk
                        last_b = bi
                else:
                    break
            if last_vk == -1:
                offset += xs.char_byte_num(xs.char_idx_of_byte[offset])
                if sep is not None:
                    while offset < xs.byte_num and xs.chars[xs.char_idx_of_byte[offset]] not in sep:
                        offset += xs.char_byte_num(xs.char_idx_of_byte[offset])
                continue
            yield last_vk, ignore_case_offset(align, xs, offset), ignore_case_offset(align, xs, last_b) + 1
            offset = last_b + 1

    def replace_longest(self, unicode s not None, callback not None, sep = None):
        """
        replace trie's keys from given string. only replace the longest.
        Args:
            s : unicode
            callback : lambda | list | dict
            sep : set(int) | None
                If you specify seperators. e.g. set([ord(' ')]), 
                it only matches strings tween seperators.
        Returns:
            replaced text
        Examples:
            >>> python_id = trie.insert("python")
            >>> text = trie.replace_longest("python", {python_id: "hahah"}, set([ord(" ")]))
        """
        cdef stringbuf sb
        cdef ignore_case_alignment align = None
        cdef xstring xs = xstring(s)
        cdef xstring prev_xs = xs
        if self.ignore_case:
            align = ignore_case_alignment(xs)
            xs = align.lowercase
        cdef int offset = 0 # byte offset
        cdef int node = 0
        cdef int bi
        cdef int last_vk = -1
        cdef int last_b = 0
        cdef int prev_offset, prev_offset2, offset2
        cdef byte_t *byte_code
        cdef bytes encoded_replaced
        cdef int char_byte_num
        cdef bool callback_list_or_dict = isinstance(callback, dict) or isinstance(callback, list)

        while offset < xs.byte_num:
            last_vk = -1
            last_b = 0
            if sep is not None:
                prev_offset = offset
                while offset < xs.byte_num and xs.chars[xs.char_idx_of_byte[offset]] in sep:
                    offset += xs.char_byte_num(xs.char_idx_of_byte[offset])
                if offset > prev_offset:
                    prev_offset2 = ignore_case_byte_index_mapping(align, prev_offset)
                    offset2 = ignore_case_byte_index_mapping(align, offset)
                    sb.write(<char*>prev_xs.bytes_ + prev_offset2, offset2 - prev_offset2)
                if offset >= xs.byte_num:
                    break
            node = 0
            for bi in range(offset, xs.byte_num):
                b = xs.bytes_[bi]
                node = self.jump(b, node)
                if node >= 0:
                    vk = self.value(node)
                    if vk >= 0:
                        if sep is not None:
                            if bi + 1 < xs.byte_num and xs.chars[xs.char_idx_of_byte[bi + 1]] not in sep:
                                continue
                        last_vk = vk
                        last_b = bi
                else:
                    break
            if last_vk == -1:
                prev_offset = offset
                offset += xs.char_byte_num(xs.char_idx_of_byte[offset])
                if sep is not None:
                    while offset < xs.byte_num and xs.chars[xs.char_idx_of_byte[offset]] not in sep:
                        offset += xs.char_byte_num(xs.char_idx_of_byte[offset])
                prev_offset2 = ignore_case_byte_index_mapping(align, prev_offset)
                offset2 = ignore_case_byte_index_mapping(align, offset)
                sb.write(<char*>prev_xs.bytes_ + prev_offset2, offset2 - prev_offset2)
                continue
            if callback_list_or_dict:
                replaced_ = callback[last_vk]
            else:
                replaced_ = callback(last_vk, ignore_case_offset(align, xs, offset), ignore_case_offset(align, xs, last_b) + 1)
            if isinstance(replaced_, unicode):
                encoded_replaced = replaced_.encode("utf8")
                byte_code = encoded_replaced
            elif isinstance(replaced_, bytes):
                encoded_replaced = replaced_
                byte_code = replaced_
            else:
                raise Exception("Replaced result should be bytes or unicode")
            sb.write(<char*>byte_code, len(encoded_replaced))
            offset = last_b + 1
        return sb.to_string().decode("utf8")

    def _dump_array(self, fname):
        """
        Used for debug
        """
        cdef Node* n
        cdef int i
        with open(fname, "w") as fo:
            for i in range(self.array_size):
                n = &self.array[i]
                fo.write("value=%s, check=%s, flags=%s, child=%s, sibling=%s\n" % (n.value, n.check, n.flags, n.child, n.sibling))

    def __reduce__(self):
        return (new_object, (Trie,), self.__getstate__())

    def __getstate__(self):
        return (self.key_num, self.key_capacity, 
            self.bheadF, self.bheadC, self.bheadO, 
            self.array_size, self.capacity, self.ordered, self.ignore_case,
            self.max_trial, self.leaf_size,
            array_to_bytes(<char*>self.array, self.array_size * sizeof(Node)),
            array_to_bytes(<char*>self.blocks, (self.array_size >> 8) * sizeof(Block)),
            array_to_bytes(<char*>self.leafs, self.leaf_size * sizeof(int)),
            array_to_bytes(<char*>self.reject, 257 * sizeof(int))
        )
    
    def __setstate__(self, data):
        self.key_num, self.key_capacity, \
        self.bheadF, self.bheadC, self.bheadO, \
        self.array_size, self.capacity, self.ordered, self.ignore_case, \
        self.max_trial, self.leaf_size, \
        array, blocks, leafs, reject = data
        if self.array != NULL:
            free(self.array)
        if self.blocks != NULL:
            free(self.blocks)
        if self.leafs != NULL:
            free(self.leafs)
        self.array = <Node*> bytes_to_array(array, (self.capacity * sizeof(Node)))
        self.blocks = <Block*> bytes_to_array(blocks, (self.capacity >> 8) * sizeof(Block))
        self.leafs = <int*> bytes_to_array(leafs, self.key_capacity * sizeof(int))
        cdef int i
        for i in range(257):
            self.reject[i] = reject[i]

    cdef write(self, FILE* ptr_fw):
        cdef uint32_t magic = magic_number
        fwrite(<void*>&magic, sizeof(magic), 1, ptr_fw)
        cdef int size = self.buff_size()
        fwrite(&size, sizeof(int), 1, ptr_fw)
        fwrite(<void*>&self.key_num, sizeof(int), 1, ptr_fw)
        fwrite(<void*>&self.key_capacity, sizeof(int), 1, ptr_fw)
        fwrite(<void*>&self.bheadF, sizeof(int), 1, ptr_fw)
        fwrite(<void*>&self.bheadC, sizeof(int), 1, ptr_fw)
        fwrite(<void*>&self.bheadO, sizeof(int), 1, ptr_fw)
        fwrite(<void*>&self.array_size, sizeof(int), 1, ptr_fw)
        fwrite(<void*>&self.capacity, sizeof(int), 1, ptr_fw)
        # for arm processor, we should align data in 4bytes.
        cdef int val = self.ordered
        fwrite(<void*>&val, sizeof(int), 1, ptr_fw)
        val = self.ignore_case
        fwrite(<void*>&val, sizeof(int), 1, ptr_fw)
        fwrite(<void*>&self.max_trial, sizeof(int), 1, ptr_fw)
        fwrite(<void*>&self.leaf_size, sizeof(int), 1, ptr_fw)
        
        fwrite(<void*>self.array, sizeof(Node), self.capacity, ptr_fw)
        fwrite(<void*>self.blocks, sizeof(Block), self.capacity >> 8, ptr_fw)
        fwrite(<void*>self.leafs, sizeof(int), self.key_capacity, ptr_fw)
        fwrite(<void*>&self.reject, sizeof(int), 257, ptr_fw)

    def buff_size(self):
        """
        return the memory size of buffer needed for exporting to external buffer.
        """
        return sizeof(uint32_t) +  sizeof(int) + sizeof(self.key_num) + sizeof(self.key_capacity) + sizeof(self.bheadF) + sizeof(self.bheadC) + sizeof(self.bheadO) + \
            sizeof(self.array_size) + sizeof(self.capacity) + sizeof(int) + sizeof(int) + sizeof(self.max_trial) + sizeof(self.leaf_size) + \
            sizeof(Node) * self.capacity + sizeof(Block) * (self.capacity >> 8) + sizeof(int) * self.key_capacity + sizeof(int) * 257


    def save(self, fname):
        """
        save data into binary file
        """
        cdef FILE *ptr_fw
        cdef bytes bfname = fname.encode("utf8")
        cdef char* path = bfname
        ptr_fw = fopen(path, "wb")
        if ptr_fw==NULL:
            raise Exception("Cannot open file: %s" % fname)
        self.write(ptr_fw)
        fclose(ptr_fw)

    cdef void _to_buff(self, void* buf):
        cdef int offset = 0

        cdef char* buff = <char*>buf
        cdef uint32_t magic = magic_number
        memcpy(buff, <void*>&magic, sizeof(magic))
        offset += sizeof(magic)

        cdef int size = self.buff_size()
        memcpy(buff + offset, &size, sizeof(int))
        offset += sizeof(int)

        memcpy(buff + offset, <void*>&self.key_num, sizeof(int))
        offset += sizeof(int)

        memcpy(buff + offset, <void*>&self.key_capacity, sizeof(int))
        offset += sizeof(int)

        memcpy(buff + offset, <void*>&self.bheadF, sizeof(int))
        offset += sizeof(int)

        memcpy(buff + offset, <void*>&self.bheadC, sizeof(int))
        offset += sizeof(int)

        memcpy(buff + offset, <void*>&self.bheadO, sizeof(int))
        offset += sizeof(int)

        memcpy(buff + offset, <void*>&self.array_size, sizeof(int))
        offset += sizeof(int)

        memcpy(buff + offset, <void*>&self.capacity, sizeof(int))
        offset += sizeof(int)

        cdef int val = self.ordered
        memcpy(buff + offset, <void*>&val, sizeof(int))
        offset += sizeof(int)

        val = self.ignore_case
        memcpy(buff + offset, <void*>&val, sizeof(int))
        offset += sizeof(int)

        memcpy(buff + offset, <void*>&self.max_trial, sizeof(int))
        offset += sizeof(int)

        memcpy(buff + offset, <void*>&self.leaf_size, sizeof(int))
        offset += sizeof(int)

        memcpy(buff + offset, <void*>self.array, sizeof(Node) * self.capacity)
        offset += sizeof(Node) * self.capacity

        memcpy(buff + offset, <void*>self.blocks, sizeof(Block) * (self.capacity >> 8))
        offset += sizeof(Block) * (self.capacity >> 8)

        memcpy(buff + offset, <void*>self.leafs, sizeof(int) * self.key_capacity)
        offset += sizeof(int) * self.key_capacity

        memcpy(buff + offset, <void*>&self.reject, sizeof(int) * 257)
        

    def to_buff(self, buff):
        """
        copy data into buff
        Args:
            buff: object satisfy Python buff protocol
        """
        check_buffer(buff)
        cdef Py_buffer view
        if PyObject_GetBuffer(buff, &view, PyBUF_WRITABLE) != 0:
            raise Exception("cannot get writable buffer: https://docs.python.org/zh-cn/3/c-api/buffer.html")
        if self.buff_size() < view.len:
            raise Exception("buff size is smaller than needed.")
        cdef void *buf = view.buf
        self._to_buff(buf)
        PyBuffer_Release(&view)




    @classmethod
    def from_buff(cls, buff, copy = True):
        """
        init trie from buff
        Args:
            buff: object satisfy Python buff protocol https://docs.python.org/zh-cn/3/c-api/buffer.html
            copy: whether copy data, by default, it copies data from buff
        """
        check_buffer(buff)
        cdef Py_buffer* view = <Py_buffer*>malloc(sizeof(Py_buffer))
        if PyObject_GetBuffer(buff, view, PyBUF_SIMPLE) != 0:
            free(view)
            raise Exception("cannot get readable buffer: https://docs.python.org/zh-cn/3/c-api/buffer.html")
        cdef Trie trie = trie_from_buff(view.buf, view.len, copy)
        if copy:
            trie.buff = NULL
            PyBuffer_Release(view)
            free(view)
        else:
            trie.buff = view
        return trie


cdef Trie trie_from_buff(void* buf, int buf_size, bool copy):
    cdef int offset = 0
    cdef Trie trie = new_object(Trie)
    cdef uint32_t magic 
    cdef int size
    cdef char* buff = <char*>buf
    memcpy(<void*>&magic, buff, sizeof(magic))
    if magic != magic_number and magic != legacy_magic_number:
        raise Exception("invalid data, magic number is not correct")
    offset += sizeof(magic)

    memcpy(&size, buff + offset, sizeof(int))
    offset += sizeof(int)
    if size > buf_size:
        raise Exception("invalid data, buf size is not correct")

    cdef int value = 0
    memcpy(<void*>&value, buff + offset, sizeof(int))
    trie.key_num = value
    offset += sizeof(int)

    memcpy(<void*>&value, buff + offset, sizeof(int))
    trie.key_capacity = value
    offset += sizeof(int)

    memcpy(<void*>&value, buff + offset, sizeof(int))
    trie.bheadF = value
    offset += sizeof(int)

    memcpy(<void*>&value, buff + offset, sizeof(int))
    trie.bheadC = value
    offset += sizeof(int)

    memcpy(<void*>&value, buff + offset, sizeof(int))
    trie.bheadO = value
    offset += sizeof(int)

    memcpy(<void*>&value, buff + offset, sizeof(int))
    trie.array_size = value
    offset += sizeof(int)

    memcpy(<void*>&value, buff + offset, sizeof(int))
    trie.capacity = value
    offset += sizeof(int)

    memcpy(<void*>&value, buff + offset, sizeof(int))
    trie.ordered = value
    offset += sizeof(int)

    memcpy(<void*>&value, buff + offset, sizeof(int))
    trie.ignore_case = value
    offset += sizeof(int)

    memcpy(<void*>&value, buff + offset, sizeof(int))
    trie.max_trial = value
    offset += sizeof(int)

    memcpy(<void*>&value, buff + offset, sizeof(int))
    trie.leaf_size = value
    offset += sizeof(int)

    if copy:
        trie.array = <Node*>malloc(sizeof(Node) * trie.capacity)
        memcpy(<void*>trie.array, buff + offset, sizeof(Node) * trie.capacity)
        offset += sizeof(Node) * trie.capacity

        trie.blocks = <Block*>malloc(sizeof(Block) * (trie.capacity >> 8))
        memcpy(<void*>trie.blocks, buff + offset, sizeof(Block) * (trie.capacity >> 8))
        offset += sizeof(Block) * (trie.capacity >> 8)

        trie.leafs = <int*>malloc(sizeof(int) * trie.key_capacity)
        memcpy(<void*>trie.leafs, buff + offset, sizeof(int) * trie.key_capacity)
        offset += sizeof(int) * trie.key_capacity
    else:
        trie.array =  <Node*>(buff + offset)
        offset += sizeof(Node) * trie.capacity

        trie.blocks = <Block*>(buff + offset)
        offset += sizeof(Block) * (trie.capacity >> 8)

        trie.leafs = <int*>(buff + offset)
        offset += sizeof(int) * trie.key_capacity


    memcpy(<void*>&trie.reject, buff + offset, sizeof(int) * 257)
    return trie










from cpython.buffer cimport PyObject_GetBuffer, PyObject_CheckBuffer, PyBuffer_Release, PyBuffer_GetPointer, Py_buffer, PyBUF_WRITABLE, PyBUF_SIMPLE

cdef void check_buffer(buff):
    if PyObject_CheckBuffer(buff) == 0:
        raise Exception("the argument doesn't support buff protocol https://docs.python.org/zh-cn/3/c-api/buffer.html")


#-*- coding:utf-8 -*- 
import unittest
import pickle
from cyac import Trie, AC

class TestPickle(unittest.TestCase):
    def test_pickle_trie(self):
        trie = Trie(ignore_case=True)
        ids = {w : trie.insert(w) for w in [u"aİİ", u"aai̇", u"aai̇bİ"]}
        with open("trie.pkl", "wb") as fo:
            pickle.dump(trie, fo)
        with open("trie.pkl", "rb") as fi:
            trie = pickle.load(fi)
        replaced = {
            ids[u"aİİ"] : u"a",
            ids[u"aai̇"] : u"b",
            ids[u"aai̇bİ"] : u"c",
        }
        res = trie.replace_longest(u"aaİ aai̇bİaa", lambda x, start, end: replaced[x])
        self.assertEqual(res, u"b caa")
        sep = set([ord(" ")]) # space as seperator
        res = trie.replace_longest(u"aaİ aai̇bİaa", lambda x, start, end: replaced[x], sep)
        self.assertEqual(res, u"b aai̇bİaa")

    def test_pickle_ac(self):
        ac = AC.build([u"aİ", u"aaİ", u"aai̇", u"aai̇bİ"], True)
        with open("ac.pkl", "wb") as fo:
            pickle.dump(ac, fo)
        with open("ac.pkl", "rb") as fi:
            ac = pickle.load(fi)
        self.assertEqual(ac.size, 3)
        arr = [(end_, val) for val, start_, end_ in ac.match(u"aai̇bİa")]
        self.assertEqual(arr, [(4, 1), (4, 0), (6, 2)])


# distutils: language=c++
from libc.string cimport strlen, memcpy, memset
from libc.stdlib cimport malloc, free, realloc
from libcpp cimport bool
from libcpp.deque  cimport deque
from libc.stdio cimport *
from libcpp cimport bool
import cython
from .xstring cimport xstring, byte_t, ignore_case_alignment, unicode_int_t, stringbuf


cdef struct Block:
    int prev
    int next_
    int num
    int reject
    int trial
    int ehead


cdef struct Node:
    int value
    int check
    byte_t sibling
    byte_t child
    unsigned short flags

cpdef inline bytes array_to_bytes(char* ptr, int size):
    return <bytes>ptr[:size]

cpdef inline char* bytes_to_array(bytes data, int capacity):
    cdef char* ptr = <char*> malloc(capacity)
    memcpy(ptr, <char *>data, len(data))
    return ptr


cdef inline int ignore_case_offset(ignore_case_alignment align, xstring xs, int byte_idx):
    cdef int char_offset = xs.char_idx_of_byte[byte_idx]
    if align is None:
        return char_offset
    return align.lowercase_char_index_mapping[char_offset]

cdef class Trie(object):
    cdef int key_num
    cdef int key_capacity
    cdef Node* array
    cdef Block* blocks
    cdef int reject[257]
    cdef int bheadF
    cdef int bheadC
    cdef int bheadO
    cdef int array_size
    cdef int capacity
    cdef bool ignore_case
    cdef bool ordered
    cdef int max_trial
    cdef int last_remove_leaf
    cdef int* leafs
    cdef int leaf_size
    cdef Py_buffer* buff
    cdef inline int _get(self, byte_t *key, int key_size, int from_, int start)
    cdef inline int follow(self, int from_, byte_t label)
    cdef bool has_label(self, int id_, byte_t label)
    cdef inline int child(self, int id_, byte_t label)
    cdef inline int children(self, int id_, byte_t *labels, int *children_arr, int first_n)
    cdef inline int jump(self, byte_t byte, int from_)
    cdef inline int jump_bytes(self, byte_t *bytes_, int byte_num, int from_)
    cdef inline int jump_uchar(self, xstring text, int uchar_idx, int from_)
    cdef bytes substring(self, int id_, int start_id)
    cdef bytes key(self, int id_)
    cdef inline int value(self, int id_)
    cdef inline bool has_value(self, int id_)
    cpdef int insert(self, unicode key)
    cpdef int remove(self, unicode key_)
    cpdef int get(self, unicode key)
    cdef inline int get_bytes(self, byte_t* bkey, int len_)

    # Private Method, Don't Call
    cdef inline void _node_init(self, Node* n, int value, int check)
    cdef inline int _node_child_num(self, Node *n)
    cdef inline void _node_set_child_num(self, Node *n, int val)
    cdef inline int _node_base(self, Node *n)
    cdef inline bool _node_is_child(self, Node *n, int par)
    cdef inline byte_t* _node_child_ptr(self, Node *n)
    cdef inline byte_t* _node_sibling_ptr(self, Node *n)
    cdef inline void _block_init(self, Block* block, int prev, int next_, int trial, int ehead, int num, int reject)
    cdef inline Node* _node(self, int nid)
    cdef inline void pop_block(self, int bi, int* head_in, bool last)
    cdef inline void push_block(self, int bi, int* head_out, bool empty)
    cdef int add_block(self)
    cdef inline void transfer_block(self, int bi, int* head_in, int* head_out)
    cdef inline int pop_enode(self, int base, byte_t label, int from_)
    cdef void push_enode(self, int e)
    cdef inline void push_sibling(self, int from_, int base, byte_t label, bool has_child)
    cdef inline void pop_sibling(self, int from_, byte_t label)
    cdef inline bool consult(self, Node *nref, Node *pref)
    cdef inline int sibling(self, int to)
    cdef int set_child(self, int base, byte_t c, byte_t label, bool append_label, byte_t *children)
    cdef inline int find_place(self)
    cdef inline int find_places(self, byte_t *child, int child_num)
    cdef inline int resolve(self, int from_n, int base_n, byte_t label_n)
    cdef void _to_buff(self, void* buff)
    cdef write(self, FILE* ptr_fw)

cdef Trie trie_from_buff(void* buf, int buf_size, bool copy)



#-*- coding:utf-8 -*- 
import unittest
from cyac import Trie, AC
import sys

class TestBuff(unittest.TestCase):
    def test_buff_ac(self):
        trie = Trie(ignore_case=True)
        ids = {w : trie.insert(w) for w in [u"aİİ", u"aai̇", u"aai̇bİ"]}
        trie.save("trie.bin")
        with open("trie.bin", "rb") as fi:
            bs = bytearray(fi.read())
        self.assertEqual(len(bs), trie.buff_size())
        bs2 = bytearray(trie.buff_size())
        trie.to_buff(bs2)
        self.assertEqual(bs2, bs)

        self._check_trie_correct(Trie.from_buff(bs2, copy=True), ids)
        self._check_trie_correct(Trie.from_buff(bs2, copy=False), ids)


    def _check_trie_correct(self, trie, ids):
        replaced = {
            ids[u"aİİ"] : u"a",
            ids[u"aai̇"] : u"b",
            ids[u"aai̇bİ"] : u"c",
        }
        res = trie.replace_longest(u"aaİ aai̇bİaa", lambda x, start, end: replaced[x])
        self.assertEqual(res, u"b caa")
        sep = set([ord(" ")]) # space as seperator
        res = trie.replace_longest(u"aaİ aai̇bİaa", lambda x, start, end: replaced[x], sep)
        self.assertEqual(res, u"b aai̇bİaa")

    def test_buff_ac(self):
        ac = AC.build([u"aİ", u"aaİ", u"aai̇", u"aai̇bİ"], True)
        ac.save("ac.bin")
        with open("ac.bin", "rb") as fi:
            bs = bytearray(fi.read())
        self.assertEqual(len(bs), ac.buff_size())
        bs2 = bytearray(ac.buff_size())
        ac.to_buff(bs2)
        self.assertEqual(bs2, bs)
        self._check_ac_correct(ac)
        self._check_ac_correct(AC.from_buff(bs2, copy=True))
        self._check_ac_correct(AC.from_buff(bs2, copy=False))

    def _check_ac_correct(self, ac):
        self.assertEqual(ac.size, 3)
        arr = [(end_, val) for val, start_, end_ in ac.match(u"aai̇bİa")]
        self.assertEqual(arr, [(4, 1), (4, 0), (6, 2)])

if __name__ == '__main__':
    unittest.main()



# distutils: language=c++
ctypedef unsigned char byte_t
cdef inline byte_t* iter_unicode(byte_t *src, int* ret, int *char_byte_num):
    cdef int result = 0
    cdef int i
    cdef byte_t leading_byte, c
    cdef int len_
    leading_byte = src[0]
    src += 1
    if leading_byte == 0:
        ret[0] = 0
        char_byte_num[0] = 1
        return src
    if leading_byte < 0x80:
        ret[0] = leading_byte
        char_byte_num[0] = 1
        return src
    if leading_byte & 0xe0 == 0xc0:
        len_ = 2
        result = leading_byte & 0x1f
    elif leading_byte & 0xf0 == 0xe0:
        len_ = 3
        result = leading_byte & 0x0f
    elif leading_byte & 0xf8 == 0xf0:
        len_ = 4
        result = leading_byte & 0x07
    elif leading_byte & 0xfc == 0xf8:
        len_ = 5
        result = leading_byte & 0x03
    elif leading_byte & 0xfe == 0xfc:
        len_ = 6
        result = leading_byte & 0x01
    else:
        len_ = 1
        ret[0] = 0
        return NULL

    char_byte_num[0] = len_
    for i in range(len_ - 1):
        c = src[0]
        src += 1
        result <<= 6
        result += c & 0x3f
    ret[0] = result
    return src


cdef inline byte_t* encode_unicode(int c, byte_t *utf8_dst):
    if c == 0:
        return utf8_dst
    cdef byte_t ch
    if c < 0x00080:
        ch = (c & 0xFF);
        utf8_dst[0]=(ch); utf8_dst += 1
    elif c < 0x00800:
        ch = (0xC0 + ((c >> 6) & 0x1F))
        utf8_dst[0]=(ch); utf8_dst += 1
        ch = (0x80 + (c & 0x3F))
        utf8_dst[0]=(ch); utf8_dst += 1
    elif c < 0x10000:
        ch = (0xE0 + ((c >> 12) & 0x0F))
        utf8_dst[0]=(ch); utf8_dst += 1
        ch = (0x80 + ((c >> 6) & 0x3F))
        utf8_dst[0]=(ch); utf8_dst += 1
        ch = (0x80 + (c & 0x3F))
        utf8_dst[0]=(ch); utf8_dst += 1
    elif c < 0x200000:
        ch = (0xF0 + ((c >> 18) & 0x07))
        utf8_dst[0]=(ch); utf8_dst += 1
        ch = (0x80 + ((c >> 12) & 0x3F))
        utf8_dst[0]=(ch); utf8_dst += 1
        ch = (0x80 + ((c >> 6)  & 0x3F))
        utf8_dst[0]=(ch); utf8_dst += 1
        ch = (0x80 + (c & 0x3F))
        utf8_dst[0]=(ch); utf8_dst += 1
    elif c < 0x8000000:
        ch = (0xF8 + ((c >> 24) & 0x03))
        utf8_dst[0]=(ch); utf8_dst += 1
        ch = (0x80 + ((c >> 18) & 0x3F))
        utf8_dst[0]=(ch); utf8_dst += 1
        ch = (0x80 + ((c >> 12) & 0x3F))
        utf8_dst[0]=(ch); utf8_dst += 1
        ch = (0x80 + ((c >> 6)  & 0x3F))
        utf8_dst[0]=(ch); utf8_dst += 1
        ch = (0x80 + (c & 0x3F))
        utf8_dst[0]=(ch); utf8_dst += 1
    else:
        ch = (0xFC + ((c >> 30) & 0x01))
        utf8_dst[0]=(ch); utf8_dst += 1
        ch = (0x80 + ((c >> 24) & 0x3F))
        utf8_dst[0]=(ch); utf8_dst += 1
        ch = (0x80 + ((c >> 18) & 0x3F))
        utf8_dst[0]=(ch); utf8_dst += 1
        ch = (0x80 + ((c >> 12) & 0x3F))
        utf8_dst[0]=(ch); utf8_dst += 1
        ch = (0x80 + ((c >> 6)  & 0x3F))
        utf8_dst[0]=(ch); utf8_dst += 1
        ch = (0x80 + (c & 0x3F))
        utf8_dst[0]=(ch); utf8_dst += 1

    return utf8_dst


cdef int char_num(byte_t *src)
cdef int fill_char_info(byte_t *src, int *char_idx_of_byte, int *chars, int *char_offsets)




#cython: language_level=3, boundscheck=False, overflowcheck=False
#     profile=True, linetrace=True
# If return NULL then it's invalid utf8
from libc.stdlib cimport malloc, free, realloc
from cython cimport typeof
from .utf8 cimport fill_char_info, char_num
from cython cimport Py_UCS4
cdef extern from "unicode_portability.c":
    cdef extern int _PyUnicode_ToLowerFull(Py_UCS4 ch, Py_UCS4* res)

# from libc.wctype cimport towlower, towupper, iswlower, iswupper


cdef class xstring(object):
    def __cinit__(self, unicode text):
        cdef bytes py_bytes = text.encode("utf8")
        self.py_unicode = text
        self.py_bytes = py_bytes
        self.bytes_ = py_bytes
        self.byte_num = len(py_bytes)
        self.char_num = char_num(self.bytes_)
        if self.char_num < 0:
            raise Exception("Invalid UTF8 string")
        self.char_idx_of_byte = <int*> malloc((self.byte_num + 1  + self.char_num + self.char_num + 1) * sizeof(int))
        self.chars = self.char_idx_of_byte + self.byte_num + 1
        self.char_offsets =self.chars + self.char_num
        fill_char_info(self.bytes_, self.char_idx_of_byte, self.chars, self.char_offsets)

    property byte_num:
        def __get__(self):
            return self.byte_num

    property char_num:
        def __get__(self):
            return self.char_num

    cpdef int char_at(self, int i):
        if i >= self.char_num:
            return 0
        return self.chars[i]

    property bytes:
        def __get__(self):
            return self.py_bytes

    def __dealloc__(self):
        if self.char_idx_of_byte:
            free(self.char_idx_of_byte)

    def __repr__(self):
        return repr(self.py_bytes)

    def __str__(self):
        return str(self.py_bytes)


cdef class ignore_case_alignment(object):
    def __cinit__(self, xstring original):
        self.original = original
        self.lowercase = xstring(original.py_unicode.lower())
        self.lowercase_char_index_mapping =  <int*> malloc((len(self.lowercase.py_unicode) + 1) * sizeof(int))
        cdef Py_UCS4 buf[4]
        cdef int offset = 0
        cdef int n
        for ori_char_idx in range(len(original.py_unicode)):
            n = _PyUnicode_ToLowerFull(original.chars[ori_char_idx], buf)
            for i in range(n):
                self.lowercase_char_index_mapping[offset + i] = ori_char_idx
            offset += n
        self.lowercase_char_index_mapping[offset] = original.char_num

    property lowercase_xstring:
        def __get__(self):
            return self.lowercase

    property original_xstring:
        def __get__(self):
            return self.original



    def alignment_array(self):
        return [self.lowercase_char_index_mapping[cidx] for cidx in range(len(self.lowercase.py_unicode))]


    def __dealloc__(self):
        if self.lowercase_char_index_mapping:
            free(self.lowercase_char_index_mapping)




#-*- coding:utf-8 -*- 
import unittest, sys
import os
from cyac import Trie

class TestTrie(unittest.TestCase):
    def test_init(self):
        trie = Trie()
        ids = [trie.insert(w) for w in [u'Ruby', u'ruby', u'rb']]
        self.assertEqual(ids, [0, 1, 2])
        self.assertEqual(u"ruby" in trie, True)
        self.assertEqual(u"rubyx" in trie, False)
        self.assertEqual(trie.remove(u"ruby"), 1)
        self.assertEqual(trie.remove(u"ruby"), -1)

    def test_get_element(self):
        trie = Trie()
        ids = {w : trie.insert(w) for w in [u"ruby", u"rubx", u"rab", u"rub", u"rb"]}
        for w, id_ in ids.items():
            self.assertEqual(trie[id_], w)
            self.assertEqual(trie[w], id_)
    
    def test_init2(self):
        trie = Trie()
        ids = [trie.insert(w) for w in [u'Ruby', u'ruby', u'rb', u'XX']]
        self.assertEqual([(k, v) for k, v in trie.items()], [(u"Ruby", 0), (u"ruby", 1), (u"rb", 2), (u"XX", 3)])

    def test_ignore_case(self):
        trie = Trie(ignore_case=True)
        ids = [trie.insert(w) for w in [u'Ruby', u'ruby', u'rb']]
        self.assertEqual(ids, [0, 0, 1])
        self.assertEqual(trie.remove(u"ruby"), 0)
        self.assertEqual(trie.remove(u"Ruby"), -1)

    def test_prefix(self):
        trie = Trie()
        ids = {w : trie.insert(w) for w in [u"ruby", u"rubx", u"rab", u"rub"]}
        prefixes = list(trie.prefix(u"ruby on rails"))
        self.assertEqual(prefixes, [(ids[u"rub"], 3), (ids[u"ruby"], 4)])

    def test_ignore_case_prefix(self):
        if sys.version_info.major < 3:
            return
        txt = u"aaİbİc"
        trie = Trie(ignore_case=True)
        ids = {w : trie.insert(w) for w in [u"aİ", u"aİİ", u"aai̇", u"aai̇bİ"]}
        prefixes = list(trie.prefix(txt))
        self.assertEqual(prefixes, [(ids[u"aai̇"], 3), (ids[u"aai̇bİ"], 5)])
        txt = u"aai̇bİc"
        prefixes = list(trie.prefix(txt))
        self.assertEqual(prefixes, [(ids[u"aai̇"], 4), (ids[u"aai̇bİ"], 6)])

    def test_predict(self):
        trie = Trie(ordered=True)
        ids = {w : trie.insert(w) for w in [u"ruby", u"rubx", u"rab", u"rub", u"rb"]}
        predicts = list(trie.predict(u"r"))
        self.assertEqual(predicts, [
            ids[u"rb"], 
            ids[u"rab"], 
            ids[u"rub"], 
            ids[u"rubx"], 
            ids[u"ruby"]])
    
    def test_ignore_case_predict(self):
        if sys.version_info.major < 3:
            return
        trie = Trie(ignore_case=True, ordered=True)
        ids = {w : trie.insert(w) for w in [u"aaİ", u"aİİ", u"aai̇", u"aai̇bİ"]}
        predicts = list(trie.predict(u"aaİ"))
        self.assertEqual(predicts, [
            ids[u'aai̇'],
            ids[u'aai̇bİ']
        ])

    def test_match_longest(self):
        trie = Trie()
        ids = {w : trie.insert(w) for w in [u"New York", u"New", u"York", u"York City", u"City", u"City is"]}
        matches = list(trie.match_longest(u"New York City isA"))
        self.assertEqual(matches, [
            (ids[u"New York"], 0, len(u"New York")),
            (ids[u"City is"], len(u"New York "), len(u"New York City is"))
        ])
        sep = set([ord(" ")]) # space as seperator
        matches = list(trie.match_longest(u"New York City isA", sep))
        self.assertEqual(matches, [
            (ids[u"New York"], 0, len(u"New York")),
            (ids[u"City"], len(u"New York "), len(u"New York City"))
        ])

    def test_ignore_case_match_longest(self):
        if sys.version_info.major < 3:
            return
        trie = Trie(ignore_case=True)
        ids = {w : trie.insert(w) for w in [u"aİİ", u"aai̇", u"aai̇bİ"]}
        matches = list(trie.match_longest(u"aaİ aai̇bİaa"))
        self.assertEqual(matches, [
            (ids[u"aai̇"], 0, len(u"aaİ")),
            (ids[u"aai̇bİ"], len(u"aaİ "), len(u"aaİ aai̇bİ"))
        ])
        sep = set([ord(" ")]) # space as seperator
        matches = list(trie.match_longest(u"aaİ aai̇bİaa", sep))
        self.assertEqual(matches, [
            (ids[u"aai̇"], 0, len(u"aaİ")),
        ])

    def test_replace_longest(self):
        trie = Trie()
        ids = {w : trie.insert(w) for w in [u"New York", u"New", u"York", u"York City", u"City", u"City is"]}
        replaced = {
            ids[u"New York"] : u"Beijing",
            ids[u"New"] : u"Old",
            ids[u"York"] : u"Yark",
            ids[u"York City"] : u"Yerk Town",
            ids[u"City"] : u"Country",
            ids[u"City is"] : u"Province are"
        }
        res = trie.replace_longest(u"New York  City isA", lambda x, start, end: replaced[x])
        self.assertEqual(res, u"Beijing  Province areA")

        sep = set([ord(" ")]) # space as seperator
        res = trie.replace_longest(u"New York  City isA", lambda x, start, end: replaced[x], sep)
        self.assertEqual(res, u"Beijing  Country isA")

    def test_ignore_case_replace_longest(self):
        if sys.version_info.major < 3:
            return
        trie = Trie(ignore_case=True)
        ids = {w : trie.insert(w) for w in [u"aİİ", u"aai̇", u"aai̇bİ"]}
        replaced = {
            ids[u"aİİ"] : u"a",
            ids[u"aai̇"] : u"b",
            ids[u"aai̇bİ"] : u"c",
        }
        res = trie.replace_longest(u"aaİ aai̇bİaa", lambda x, start, end: replaced[x])
        self.assertEqual(res, u"b caa")
        sep = set([ord(" ")]) # space as seperator
        res = trie.replace_longest(u"aaİ aai̇bİaa", lambda x, start, end: replaced[x], sep)
        self.assertEqual(res, u"b aai̇bİaa")

    def test_reuse_id(self):
        trie = Trie()
        ids = {w : trie.insert(w) for w in [u"abc", u"abd", u"abe"]}
        trie.remove(u"abc")
        trie.remove(u"abe")
        v = trie.insert(u"abf")
        self.assertEqual(v, ids[u"abe"])
        v = trie.insert(u"abg")
        self.assertEqual(v, ids[u"abc"])
        v = trie.insert(u"abh")
        self.assertEqual(v, 3)
        v = trie.insert(u"abi")
        self.assertEqual(v, 4)

    def test_remove_words(self):
        dir_ = os.path.dirname(__file__)
        trie = Trie()
        for i in range(3):
            ids = []
            words = []
            with open(os.path.join(dir_, "../bench/words.txt")) as fi:
                for l in fi:
                    l = l.strip()
                    if isinstance(l, bytes):
                        l = l.decode("utf8")
                    if len(l) > 0:
                        words.append(l)
                        ids.append(trie.insert(l))
            for id_, w in zip(ids, words):
                self.assertEqual(id_, trie.remove(w))


    def test_match_words(self):
        dir_ = os.path.dirname(__file__)
        trie = Trie()
        ids = []
        with open(os.path.join(dir_, "../bench/words.txt")) as fi:
            for l in fi:
                l = l.strip()
                if isinstance(l, bytes):
                    l = l.decode("utf8")
                if len(l) > 0:
                    ids.append(trie.insert(l))
        with open(os.path.join(dir_, "../bench/words.txt")) as fi:
            txt = fi.read()
            if isinstance(txt, bytes):
                txt = txt.decode("utf8")
        sep = set([ord("\n")])
        matched = []
        for v, start, end in trie.match_longest(txt, sep):
            matched.append(v)
            self.assertEqual(txt[start:end], trie[v])
        self.assertEqual(matched, ids)

    def test_replace_words(self):
        dir_ = os.path.dirname(__file__)
        trie = Trie()
        ids = []
        with open(os.path.join(dir_, "../bench/words.txt")) as fi:
            for l in fi:
                l = l.strip()
                if isinstance(l, bytes):
                    l = l.decode("utf8")
                if len(l) > 0:
                    ids.append(trie.insert(l))
        with open(os.path.join(dir_, "../bench/words.txt")) as fi:
            txt = fi.read()
            if isinstance(txt, bytes):
                txt = txt.decode("utf8")
        sep = set([ord("\n")])
        ret = trie.replace_longest(txt, lambda v, start, end: str(v), sep).strip()
        self.assertEqual(ret, "\n".join([str(i) for i in ids]))

    def test_insert_zero_len_key(self):
        trie = Trie()
        self.assertEqual(trie.insert(u""), -1)

    def test_invalid_id(self):
        trie = Trie()
        trie.insert("hello")
        trie.insert("world")
        with self.assertRaises(AttributeError):
            trie[3]

        trie.remove("hello")
        with self.assertRaises(AttributeError):
            trie[0]
        
        trie.insert("hello2")
        trie[0]



#-*- coding:utf-8 -*- 
import unittest
from cyac import AC
import sys

class TestAC(unittest.TestCase):
    def test_init(self):
        ac = AC.build([u'我', u'我是', u'是中'])
        arr = [(end_, val) for val, start_, end_ in ac.match(u"我是中国人")]
        self.assertEqual(arr, [(1, 0), (2, 1), (3, 2)])

    def test_sep(self):
        ac = AC.build([u"a", u"aa", u"A", u"AA"])
        sep = set([ord(" ")])
        arr = [(end_, val) for val, _, end_ in ac.match(u"a aaa", sep)]
        self.assertEqual(arr, [(1, 0)])

    def test_ignore_case(self):
        if sys.version_info.major < 3:
            return
        ac = AC.build([u"aİ", u"aİİ", u"aai̇", u"aai̇bİ"], True)
        arr = [(end_, val) for val, start_, end_ in ac.match(u"aai̇bİa")]
        self.assertEqual(arr, [(4, 2), (4, 0), (6, 3)])

        ac = AC.build([u"aİ", u"aaİ", u"aai̇", u"aai̇bİ"], True)
        self.assertEqual(ac.size, 3)
        arr = [(end_, val) for val, start_, end_ in ac.match(u"aai̇bİa")]
        self.assertEqual(arr, [(4, 1), (4, 0), (6, 2)])

    def test_ignore_case_sep(self):
        if sys.version_info.major < 3:
            return
        ac = AC.build([u"aİ", u"aaİ", u"aai̇", u"aai̇bİ"], True)
        sep = set([ord(" ")])
        arr = [(end_, val) for val, start_, end_ in ac.match(u"aai̇bİ", sep)]
        self.assertEqual(arr, [(6, 2)])

    def test_match_all(self):
        ac = AC.build([u"hello@gmail.comhi", u"gmail.com"])
        var1 = "gmailhello@gmail.comhiaa"
        lst = list(ac.match(var1))
        self.assertEqual(len(lst), 2)

    def test_match_longest(self):
        ac = AC.build([u"hello@gmail.comhi", u"gmail.com"])
        var1 = "gmailhello@gmail.comhiaa"
        lst = list(ac.match(var1, return_all=False))
        self.assertEqual(len(lst), 1)



# distutils: language=c++

import cython
from cpython.version cimport PY_MAJOR_VERSION
from libcpp.string cimport string
from .utf8 cimport byte_t

ctypedef fused text_t:
    unicode
    bytes

ctypedef fused unicode_int_t:
    unicode
    int

cdef class xstring(object):
    cdef unicode py_unicode
    cdef bytes py_bytes
    cdef byte_t* bytes_ 
    # byte num 和 char_num 都不包括结尾的\0
    cdef int byte_num
    cdef int char_num
    cdef int* char_idx_of_byte # 每个byte对应的char
    cdef int* char_offsets # 每个char 在bytes中的offset
    cdef int* chars # char的数组
    cpdef int char_at(self, int i)

    cdef inline int char_byte_num(self, int i):
        return self.char_offsets[i+1] - self.char_offsets[i]

cdef class ignore_case_alignment(object):
    cdef xstring original
    cdef xstring lowercase
    cdef int* lowercase_char_index_mapping # char的 index 到 index 的转换

cdef extern from "<sstream>" namespace "std" nogil:
    ctypedef int streamsize
    cdef cppclass stringbuf:
        streamsize write "sputn" (const char* s, streamsize n);
        int put "sputc" (char c);
        stringbuf()
        string to_string "str" () const



# default_seperator = set([ord(c) for c in " \t\n\r'\"{}[]().,?!;<>+~#$%^&*-|\\/"])



#!/usr/bin/env python3

from setuptools import setup
from distutils.extension import Extension

# Delayed import; https://stackoverflow.com/questions/37471313/setup-requires-with-cython
try:
    from Cython.Build import cythonize
except ImportError:
     def cythonize(*args, **kwargs):
         from Cython.Build import cythonize
         return cythonize(*args, **kwargs)

# import os
# os.environ['CFLAGS'] = '-O0'
try:
    long_description = open("README.md", encoding="utf8").read()
except IOError:
    long_description = ""

setup(
    version="1.11",
    description="High performance Trie and Ahocorasick automata (AC automata) for python",
    name="cyac",
    url="https://github.com/nppoly/cyac",
    author="nppoly",
    author_email="nppoly@foxmail.com",
    packages=["cyac"],
    package_dir={'cyac': 'lib/cyac'},
    package_data={'cyac': ['*.pxd', 'unicode_portability.c', 'unicodetype_db.h']},
    include_package_data=True,
    long_description_content_type="text/markdown",
    long_description=long_description,
    install_requires=['cython>=0.29.0', 'Cython>=0.29.0'],
    setup_requires=['Cython'],
    ext_modules = cythonize([
        "lib/cyac/util.pyx",
        "lib/cyac/utf8.pyx",
        "lib/cyac/xstring.pyx",
        "lib/cyac/trie.pyx",
        "lib/cyac/ac.pyx"]),
    classifiers=[
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        "Programming Language :: Python",
        "Topic :: Text Processing",
        "Topic :: Text Processing :: Linguistic",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
    ]
)



#-*- coding:utf-8 -*- 
import unittest, sys
from cyac.xstring import xstring, ignore_case_alignment

class TestXString(unittest.TestCase):
    def test_init(self):
        txt = u"呵呵"
        s = xstring(txt)
        self.assertEqual(s.char_num, len(txt))
        self.assertEqual(s.byte_num, len(txt.encode("utf8")))
        for i, c in enumerate(txt):
            self.assertEqual(s.char_at(i), ord(c))

    def test_lowercase(self):
        if sys.version_info.major < 3:
            return
        txt = u"aaİb"
        s = xstring(txt)
        align = ignore_case_alignment(s)
        self.assertEqual(align.lowercase_xstring.bytes, txt.lower().encode("utf8"))
        self.assertEqual(align.alignment_array(), [0,1,2,2,3])

if __name__ == '__main__':
    unittest.main()


#cython: language_level=3, boundscheck=False, overflowcheck=False
#    , profile=True, linetrace=True
from .trie cimport Trie, ignore_case_alignment, ignore_case_offset, array_to_bytes, bytes_to_array, trie_from_buff
from .xstring cimport xstring
from .utf8 cimport byte_t
from .util cimport check_buffer
from libcpp.deque  cimport deque
from libc.stdlib cimport malloc, free, realloc
from libc.string cimport memcpy 
from libc.stdio cimport *
from libcpp cimport bool
from libc.string cimport memset
from libc.stdint cimport uint32_t
from libcpp.vector cimport vector
from cpython.buffer cimport PyObject_GetBuffer, PyObject_CheckBuffer, PyBuffer_Release, PyBuffer_GetPointer, Py_buffer, PyBUF_WRITABLE, PyBUF_SIMPLE
from .version import magic_number, ac_binary_version, legacy_magic_number

def new_object(obj):
    return obj.__new__(obj)

cdef struct OutNode:
    int next_
    int value

cdef struct Matched:
    int val
    int start
    int end

cdef struct QueueNode:
    int node_id
    int node_label
    int len

cdef class AC(object):
    cdef Trie trie
    # cdef OutNode* output
    cdef int* fails
    cdef unsigned int* key_lens
    cdef Py_buffer* buff

    property ignore_case:
        def __get__(self):
            return self.ignore_case
    
    property size:
        def __get__(self):
            return self.trie.key_num

    def __getitem__(self, key):
        return self.trie[key]

    def prefix(self, unicode s not None):
        """
        return the prefix of given string which is in the trie.
        Args:
            key : string
                keyword that you want to searh
        Iterates:
            prefixes : tuple(id, end_offset)
                s[:end_offset] matches id
        Examples:
            >>> for id_, offset in ac.prefix("python"):
            >>>     print(id_, offset)
        """
        for x in self.trie.prefix(s):
            yield x

    def predict(self, unicode s not None):
        """
        return the string in the trie which starts with given string
        Args:
            key : string
                keyword that you want to searh
        Iterates:
            predicts : id
        Examples:
            >>> for id_ in ac.predict("python"):
            >>>     print(id_)
        """
        for x in self.trie.predict(s):
            yield x
    
    def __iter__(self):
        for id_ in self.trie:
            yield id_

    def items(self):
        """
        return all key and id ordered by id
        Iterates:
            items : (unicode, int)
        Examples:
            >>> for key, id_ in ac.items():
            >>>     print(id_, key)
        """
        for x in self.trie.items():
            yield x

    def __contains__(self, unicode t):
        t in self.trie

    def match_longest(self, unicode s not None, sep = None):
        """
        extract trie's keys from given string. only return the longest.
        Args:
            s : unicode
            sep : set(int) | None
                If you specify seperators. e.g. set([ord(' ')]), 
                it only matches strings tween seperators.
        Iterates:
            matched: tuple(id, start_offset, end_offset)
        Examples:
            >>> for id_, start_offset, end_offset in ac.match_longest("python", set([ord(" ")])):
            >>>     print(id_, start_offset, end_offset)
        """
        for x in self.trie.match_longest(s, sep):
            yield x

    def replace_longest(self, unicode s not None, callback not None, sep = None):
        """
        replace trie's keys from given string. only replace the longest.
        Args:
            s : unicode
            callback : lambda | list | dict
            sep : set(int) | None
                If you specify seperators. e.g. set([ord(' ')]), 
                it only matches strings tween seperators.
        Returns:
            replaced text
        Examples:
            >>> python_id = ac.get("python")
            >>> text = ac.replace_longest("python", {python_id: "hahah"}, set([ord(" ")]))
        """
        return self.trie.replace_longest(s, callback, sep)

    def __cinit__(self):
        self.trie = None
        # self.output = NULL
        self.fails = NULL
        self.key_lens = NULL
        self.buff = NULL

    def __dealloc__(self):
        if self.buff == NULL:
            # if self.output:
            #     free(self.output)
            if self.fails:
                free(self.fails)
            if self.key_lens:
                free(self.key_lens)
        else:
            PyBuffer_Release(self.buff)
            free(self.buff)

    # cdef inline void __fetch(self, int idx, int nid, vector[Matched]& res):
    #     cdef OutNode *e = &self.output[nid]
    #     cdef int val, len_, start_offset, end_offset
    #     while e[0].value >= 0:
    #         val = e[0].value
    #         len_ = self.key_lens[val]
    #         start_offset = idx - len_ + 1
    #         end_offset = idx + 1
    #         res.push_back(Matched(val, start_offset, end_offset))
    #         if e[0].next_ < 0:
    #             break
    #         e = &self.output[e[0].next_]

    @classmethod
    def build(cls, pats, ignore_case=False, ordered=False):
        """
        Build AC automata
        Args:
            pats : list(unicode)
            ignore_case : bool
                Defaults False
                see Trie's constructor for details.
            ordered : bool
                Defaults False
                see Trie's constructor for details.
        Returns:
            ac : AC
                ac automata
        Examples:
            >>> AC.build(["python"])
        """
        cdef int idx, id_, l, vk, fid, nid, fs
        cdef byte_t label
        cdef Trie trie = Trie(ignore_case, ordered)
        for s in pats:
            trie.insert(s)
        cdef QueueNode queue_node, queue_node2
        cdef int nlen = trie.array_size
        cdef int* fails = <int*> malloc(sizeof(int)*nlen)
        memset(fails, -1, sizeof(int)*nlen)
        # cdef OutNode* output = <OutNode*> malloc(sizeof(OutNode)*nlen)
        # memset(output, -1, sizeof(OutNode)*nlen)
        cdef deque[QueueNode] q = deque[QueueNode]()
        key_lens = <unsigned int*> malloc(sizeof(unsigned int) * trie.leaf_size)
        memset(key_lens, 0, sizeof(unsigned int) * trie.leaf_size)
        cdef int ro = 0
        fails[ro] = ro
        cdef byte_t children[256]
        cdef int children_nodes[256]
        child_num = trie.children(ro, children, children_nodes, 256)
        for idx in range(child_num):
            label = children[idx]
            id_ = children_nodes[idx]
            # print("child[%s][%s] = %s" % (ro, label, id_))
            queue_node.node_id = id_
            queue_node.node_label = label
            queue_node.len = 1
            q.push_back(queue_node)
            fails[id_] = ro
        while q.size() > 0:
            queue_node2 = q.front()
            q.pop_front()
            # print("queue_node2", queue_node2)
            nid = queue_node2.node_id
            l = queue_node2.len
            vk = trie.value(nid)
            if vk >= 0:
                key_lens[vk] = l
                # print("output[%s].value = %s, label: %s" % (nid, vk, queue_node2.node_label))
                # output[nid].value = vk
            child_num = trie.children(nid, children, children_nodes, 256)
            for idx in range(child_num):
                # print("child_idx %d of %d" % (idx, nid))
                label = children[idx]
                id_ = children_nodes[idx]
                # print("child[%s][%s] = %s" % (nid, label, id_))
                queue_node.node_id = id_
                queue_node.node_label = label
                queue_node.len = l + 1
                q.push_back(queue_node)
                fid = nid
                while fid != ro:
                    fs = fails[fid]
                    if trie.has_label(fs, label):
                        fid = trie.child(fs, label)
                        break
                    fid = fails[fid]
                fails[id_] = fid
                # if trie.has_value(fid):
                #     output[id_].next_ = fid
                #     # print("output[%s].next_ = %s" % (id_, fid))
        ac = AC()
        ac.trie = trie
        # ac.output = output
        ac.fails = fails
        ac.key_lens = key_lens
        return ac

    cdef inline void __fetch(self, int idx, int nid, vector[Matched]& res, bool return_all):
        cdef int val, len_, start_offset, end_offset
        while nid > 0:
            val = self.trie.value(nid)
            if val >= 0:
                len_ = self.key_lens[val]
                start_offset = idx - len_ + 1
                end_offset = idx + 1
                res.push_back(Matched(val, start_offset, end_offset))
            if not return_all:
                break
            nid = self.fails[nid]

    def match(self, unicode text not None, sep = None, return_all = True):
        """
        extract trie's keys from given string. 
        Args:
            text : unicode
            sep : set(int) | None
                If you specify seperators. e.g. set([ord(' ')]), 
                it only matches strings between seperators.
            return_all: if it's false, only return the longest substring in substrings with same suffix. it's useful only when sep is None
        Iterates:
            matched: tuple(id, start_offset, end_offset)
        Examples:
            >>> for id_, start_offset, end_offset in ac.match_longest("python", set([ord(" ")])):
            >>>     print(id_, start_offset, end_offset)
        """
        cdef xstring xstr = xstring(text)
        cdef xstring prev_xstr = xstr
        cdef ignore_case_alignment align = None
        cdef vector[Matched] vect
        cdef Matched m
        if self.trie.ignore_case:
            align = ignore_case_alignment(xstr)
            xstr = align.lowercase
        cdef byte_t *bytes_ = xstr.bytes_
        cdef int byte_num = xstr.byte_num
        cdef int nid_, i, chr_
        cdef int nid = 0
        cdef byte_t b
        cdef bool return_all_ = return_all

        if sep is None:
            for i in range(byte_num):
                b = bytes_[i]
                while True:
                    nid_ = self.trie.child(nid, b)
                    if nid_ >= 0:
                        nid = nid_
                        vect.clear()
                        self.__fetch(i, nid, vect, return_all_)
                        for m in vect:
                            yield m.val, ignore_case_offset(align, xstr, m.start), ignore_case_offset(align, xstr, m.end - 1) + 1
                        break
                    if nid == 0:
                        break
                    nid = self.fails[nid]
        else:
            for i in range(byte_num):
                b = bytes_[i]
                while True:
                    nid_ = self.trie.child(nid, b)
                    if nid_ >= 0:
                        nid = nid_
                        if i + 1 < byte_num:
                            chr_ = xstr.chars[xstr.char_idx_of_byte[i + 1]]
                            if chr_ not in sep:
                                break
                        vect.clear()
                        self.__fetch(i, nid, vect, return_all_)
                        for m in vect:
                            if m.start > 0:
                                chr_ = xstr.chars[xstr.char_idx_of_byte[m.start - 1]]
                                if chr_ not in sep:
                                    continue
                            yield m.val, ignore_case_offset(align, xstr, m.start), ignore_case_offset(align, xstr, m.end - 1) + 1
                        break
                    if nid == 0:
                        break
                    nid = self.fails[nid]

    def __reduce__(self):
        return (new_object, (AC,), self.__getstate__())

    def __getstate__(self):
        return (self.trie,
            int(ac_binary_version),
            array_to_bytes(<char*>self.fails, self.trie.array_size * sizeof(int)),
            array_to_bytes(<char*>self.key_lens, self.trie.leaf_size * sizeof(unsigned int)),
        )

    def __setstate__(self, data):
        self.trie, version, fails, key_lens = data
        if isinstance(version, int) and version > ac_binary_version:
            raise Exception("data is generated by new cyac, please update cyac module.")
        # if self.output != NULL:
        #     free(self.output)
        if self.fails != NULL:
            free(self.fails)
        if self.key_lens != NULL:
            free(self.key_lens)
        # self.output = <OutNode*> bytes_to_array(output, self.trie.array_size * sizeof(OutNode))
        self.fails = <int*> bytes_to_array(fails, self.trie.array_size * sizeof(int))
        self.key_lens = <unsigned int*> bytes_to_array(key_lens, self.trie.leaf_size * sizeof(unsigned int))


    cdef write(self, FILE* ptr_fw):
        cdef uint32_t magic = magic_number
        cdef int version = ac_binary_version
        fwrite(<void*>&magic, sizeof(magic), 1, ptr_fw)
        fwrite(<void*>&version, sizeof(version), 1, ptr_fw)
        cdef int size = self.buff_size()
        fwrite(<void*>&size, sizeof(size), 1, ptr_fw)
        self.trie.write(ptr_fw)
        # fwrite(<void*>self.output, sizeof(OutNode), self.trie.array_size, ptr_fw)
        fwrite(<void*>self.fails, sizeof(int), self.trie.array_size, ptr_fw)
        fwrite(<void*>self.key_lens, sizeof(unsigned int), self.trie.leaf_size, ptr_fw)

    def save(self, fname):
        """
        save data into binary file
        """
        cdef FILE *ptr_fw
        cdef bytes bfname = fname.encode("utf8")
        cdef char* path = bfname
        ptr_fw = fopen(path, "wb")
        if ptr_fw==NULL:
            raise Exception("Cannot open file: %s" % fname)
        self.write(ptr_fw)
        fclose(ptr_fw)


    def buff_size(self):
        """
        return the memory size of buffer needed for exporting to external buffer.
        """
        return self.trie.buff_size() + sizeof(uint32_t) + sizeof(int) + sizeof(int) + (sizeof(int)) * self.trie.array_size + sizeof(unsigned int) * self.trie.leaf_size


    def to_buff(self, buff):
        """
        copy data into buff
        Args:
            buff: object satisfy Python buff protocol
        """
        check_buffer(buff)
        cdef Py_buffer view
        if PyObject_GetBuffer(buff, &view, PyBUF_WRITABLE) != 0:
            raise Exception("cannot get writable buffer: https://docs.python.org/zh-cn/3/c-api/buffer.html")
        if self.buff_size() < view.len:
            raise Exception("buff size is smaller than needed.")
        cdef void* buf = view.buf
        self._to_buff(buf)
        PyBuffer_Release(&view)


    @classmethod
    def from_buff(cls, buff, copy = True):
        """
        init ac from buff
        Args:
            buff: object satisfy Python buff protocol https://docs.python.org/zh-cn/3/c-api/buffer.html
            copy: whether copy data, by default, it copies data from buff
        """
        check_buffer(buff)
        cdef Py_buffer* view = <Py_buffer*>malloc(sizeof(Py_buffer))
        cdef Py_buffer* view2 = <Py_buffer*>malloc(sizeof(Py_buffer))
        if PyObject_GetBuffer(buff, view, PyBUF_SIMPLE) != 0 or PyObject_GetBuffer(buff, view2, PyBUF_SIMPLE) != 0:
            free(view)
            free(view2)
            raise Exception("cannot get readable buffer: https://docs.python.org/zh-cn/3/c-api/buffer.html")
        ac = ac_from_buff(view.buf, view.len, copy)
        if copy:
            ac.buff = NULL
            ac.trie.buff = NULL
            PyBuffer_Release(view)
            PyBuffer_Release(view2)
            free(view)
            free(view2)
        else:

            ac.buff = view
            ac.trie.buff = view2
        return ac

    cdef void _to_buff(self, void* buf):
        cdef int offset = 0

        cdef char* buff = <char*>buf
        cdef uint32_t magic = magic_number
        memcpy(buff, <void*>&magic, sizeof(magic))
        offset += sizeof(magic)

        cdef int version = ac_binary_version
        memcpy(buff + offset, <void*>&version, sizeof(version))
        offset += sizeof(version)

        cdef int size = self.buff_size()
        memcpy(buff + offset, <void*>&size, sizeof(size))
        offset += sizeof(size)

        self.trie._to_buff(buff + offset)
        offset += self.trie.buff_size()

        # memcpy(buff + offset, <void*>self.output, sizeof(OutNode) * self.trie.array_size)
        # offset += sizeof(OutNode) * self.trie.array_size

        memcpy(buff + offset, <void*>self.fails, sizeof(int) * self.trie.array_size)
        offset += sizeof(int) * self.trie.array_size

        memcpy(buff + offset, <void*>self.key_lens, sizeof(unsigned int) * self.trie.leaf_size)
        offset += sizeof(unsigned int) * self.trie.leaf_size


cdef AC ac_from_buff(void* buf, int buf_size, bool copy):

    cdef char* buff = <char*>buf
    cdef int offset = 0
    cdef AC ac = new_object(AC)
    cdef uint32_t magic
    cdef int size, ac_version
    memcpy(<void*>&magic, buff, sizeof(magic))
    if magic != magic_number and magic != legacy_magic_number:
        raise Exception("invalid data, magic number is not correct")
    offset += sizeof(magic)

    if magic == magic_number:
        memcpy(<void*>&ac_version, buff + offset, sizeof(ac_version))
        if ac_binary_version < ac_version:
            raise Exception("reading newer binary file, please update cyac")
        offset += sizeof(ac_version)

    memcpy(&size, buff + offset, sizeof(int))
    offset += sizeof(int)
    if size > buf_size:
        raise Exception("invalid data, buf size is not correct")
    trie = trie_from_buff(buff + offset, buf_size - offset, copy)
    offset += trie.buff_size()
    if copy:
        if magic == legacy_magic_number:
            offset += sizeof(OutNode) * trie.array_size
        # ac.output = <OutNode*>malloc(sizeof(OutNode) * trie.array_size)
        # memcpy(<void*>ac.output, buff + offset, sizeof(OutNode) * trie.array_size)
        # offset += sizeof(OutNode) * trie.array_size

        ac.fails = <int*>malloc(sizeof(int) * trie.array_size)
        memcpy(<void*>ac.fails, buff + offset, sizeof(int) * trie.array_size)
        offset += sizeof(int) * trie.array_size

        ac.key_lens = <unsigned int*>malloc(sizeof(unsigned int) * trie.leaf_size)
        memcpy(<void*>ac.key_lens, buff + offset, sizeof(unsigned int) * trie.leaf_size)
        offset += sizeof(unsigned int) * trie.leaf_size
    else:
        if magic == legacy_magic_number:
            offset += sizeof(OutNode) * trie.array_size
        # ac.output = <OutNode*>(buff + offset)
        # offset += sizeof(OutNode) * trie.array_size

        ac.fails = <int*>(buff + offset)
        offset += sizeof(int) * trie.array_size

        ac.key_lens = <unsigned int*>(buff + offset)
        offset += sizeof(unsigned int) * trie.leaf_size

    ac.trie = trie
    return ac



